﻿/*  Turing machine by Alexander Myasnikov
 *  e-mail: amsoftware@ya.ru
 *  Web: www.amsoftware.narod.ru
 */


using System;
using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;



namespace Turing
{

    using tcChar = System.Char;
    public delegate void debugFunc(string output);


    public class TuringCore
    {

        public class tcMoveDirection
        {
            public enum tcDirection { R, L, N };
            static public tcDirection fromString(string expression)
            {
                switch (expression)
                {
                    case "R":
                        return tcDirection.R;

                    case "L":
                        return tcDirection.L;

                    case "N":
                        return tcDirection.N;

                    default: throw new InvalidCastException();
                }

            }

        }


        private class tcRule
        {
            public tcQ whenQ { get; set; }
            public tcChar whenChar { get; set; }
            public tcQ thenQ { get; set; }
            public tcChar thenChar { get; set; }
            public tcMoveDirection.tcDirection thenDirection { get; set; }

            private tcMachine parent;

            public override string ToString()
            {
                return String.Format("Q{0}{1} → Q{2}{3}{4}", whenQ.step, whenChar, thenQ.step, thenChar, thenDirection);
            }

            public tcRule(tcMachine useParent)
            {
                parent = useParent;

            }

        }

        private class tcRuleList : List<tcRule>
        {
            // nothing to do
        };

        public class tcQ
        {
            private Dictionary<tcChar, tcRuleList> rules { get; set; }

            private tcMachine parent;

            private Int32 getStep()
            {
                return parent.IndexOf(this);
            }

            public Int32 step
            {
                get
                {
                    return getStep();
                }

            }

            public void AddRule(tcChar letter, string expression)
            {
                tcRuleList letterRules = rules.ContainsKey(letter) ? rules[letter] : rules[letter] = new tcRuleList();



                // Format = q01→q01R

                Regex re = new Regex(@"q(?<whenQ>[0-9]{1})(?<whenChar>.{1})→q(?<thenQ>[0-9]{1})(?<thenChar>.{1})(?<thenDirection>[LRN]{1})", RegexOptions.ExplicitCapture);


                MatchCollection matches = re.Matches(expression);

                foreach (Match m in matches)
                {
                    tcRule newRule = new tcRule(this.parent);

                    newRule.whenQ = parent[Convert.ToInt32(m.Groups["whenQ"].Value)];
                    newRule.whenChar = m.Groups["whenChar"].Value[0];

                    Int32 thenstep = Convert.ToInt32(m.Groups["thenQ"].Value);

                    if (parent.Count - 1 >= thenstep)
                    {
                        newRule.thenQ = parent[thenstep];
                    }
                    else
                    {
                        newRule.thenQ = parent.CreateQAtIndex(thenstep);
                    }

                    newRule.thenChar = m.Groups["thenChar"].Value[0];
                    newRule.thenDirection = tcMoveDirection.fromString(m.Groups["thenDirection"].Value);

                    letterRules.Add(newRule);
                }

            }

            public bool Process(out tcQ next)
            {

                tcRuleList letterRules = rules.ContainsKey(parent.state.Current) ? rules[parent.state.Current] : null;

                if ((letterRules != null) && (letterRules.Count != 0))
                {
                    next = null;

                    foreach (tcRule rule in letterRules)
                    {

                        if ((rule.whenQ == this) && (rule.whenChar == parent.state.Current))
                        {

                            parent.stepCounter++;

                            if (this.parent.debug != null)
                            {
                                this.parent.debug(String.Format("\rStep = {3}\rState before = {0}\rCurrent char = {2}\rActive rule = {1}", this.parent.Result, rule.ToString(), parent.state.Current, parent.stepCounter));
                            }

                            parent.state.Current = rule.thenChar;
                            parent.state.Move(rule.thenDirection);
                            next = rule.thenQ;
                            break;
                        }

                    }


                    return (next != null);
                }
                else
                {
                    next = null;
                    return false;
                }
            }

            public tcQ(tcMachine setParent)
            {
                parent = setParent;
                rules = new Dictionary<tcChar, tcRuleList>();

            }

        }

        public class tcQList : List<tcQ>
        {
            // nothing to do
        }


        public class tcState : List<tcChar>
        {

            private Int32 position { get; set; }
            private tcChar blank;

            public tcChar Current
            {

                get { return GetCurrent(); }
                set { SetCurrent(value); }

            }

            public tcState(string initial, tcChar useBlank)
                : base(initial)
            {
                position = 0;
                blank = useBlank;

            }

            public void Move(tcMoveDirection.tcDirection direction)
            {

                switch (direction)
                {
                    case tcMoveDirection.tcDirection.L:
                        position--;
                        if (position < 0)
                        {
                            this.Insert(0, this.blank);
                            position = 0;
                        }
                        break;
                    case tcMoveDirection.tcDirection.R:
                        position++;

                        if (position > this.Count - 1)
                        {
                            this.Add(this.blank);
                        }

                        break;
                    case tcMoveDirection.tcDirection.N:
                        break;

                }


            }

            private tcChar GetCurrent()
            {
                return this[position];

            }

            private void SetCurrent(tcChar value)
            {
                this[position] = value;

            }

        }



        public class tcMachine : tcQList
        {
            public tcState state;
            private tcQ activeQ;
            public Int32 stepCounter { get; set; }
            public debugFunc debug { get; set; }


            public string Result
            {
                get { return new string(state.ToArray()); }
            }

            public tcMachine(string initialState, Int32 capacity, debugFunc usedebug, tcChar useBlank)
            {
                debug = usedebug;
                stepCounter = 0;
                state = new tcState(initialState, useBlank);

                capacity = capacity == 0 ? 1 : capacity;

                for (int i = 0; i != capacity; i++)
                {
                    this.Add(new tcQ(this));
                }

                activeQ = this[0];

            }

            public void addCell(Int32 qIndex, tcChar letter, String expression)
            {

                tcQ Qitem = null;

                if (this.Count - 1 < qIndex)
                {
                    Qitem = CreateQAtIndex(qIndex);

                }
                else
                {
                    Qitem = this[qIndex];
                }

                Qitem.AddRule(letter, expression);



            }


            public tcQ CreateQAtIndex(Int32 index)
            {
                tcQ tempQ = null;

                Int32 last = this.Count - 1;

                for (int i = 0; i != index - last; i++)
                {
                    tempQ = new tcQ(this);
                    this.Add(tempQ);
                }
                return tempQ;

            }


            public bool Step()
            {

                if (activeQ != null)
                {
                    tcQ next;

                    if (activeQ.Process(out next))
                    {
                        activeQ = next;
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                else return false;
            }


        }



    }
}
